pragma solidity ^0.4.0;

/**
 * Solidity implementation of best tariff matching smart contract variant.
 * license: MIT
 * author: Christoph Leixnering
 * version: 1.2
 * date: 2017-10-15
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */
contract BestTariffVariant {
    
    /**
     * Address id of contract owner.
     * 20 byte value matching to Ethereum address.
     */
    address owner;
    
    /**
     * Hash value of embedded load forecast.
     * 32 bytes - so use at least sha256 for hashing.
     */
    bytes32 loadforecasthash;
    
    /**
     * Data structure for hashed template load profiles offered by utilities.
     * @param loadprofilehashes - byte array of hashed template load profiles.
     */
    struct Templatehash {
        bytes32[] loadprofilehashes;
    }
    
    /**
     * Array of offered embedded load profiles of type Templatehash.
     */
    mapping (address => Templatehash) templatehashes;
    
    /**
     * Byte array holding a hexadecimal representation of smart meter embedding.
     */
    bytes32[32] forecast;
    
    /**
     * Data structure for calculated distance of utility template load profile.
     * @param utility - address of utility.
     * @param distance - calculated Hamming distance.
     * @param index - index of template load profile.
     */
    struct TemplateDistance {
        address utility;
        uint16 distance;
        uint index;
    }
    
    /**
     * Array of calculated distances of type TemplateDistance.
     */
    TemplateDistance[] templateDistances;
    
    /**
     * Mapping table for Hamming distance calculation with xor.
     */
    uint8[16] map = [0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4];
    
    /**
     * Event for commit function.
     * @param ret - return message.
     * @return void
     */
    event CommitEvent(string ret);
    
    /**
     * Event for smopen function.
     * @param ret - return message.
     * @return void
     */
    event SmopenEvent(string ret);
    
    /**
     * Event for uopen function.
     * @param ret - return message.
     * @return void
     */
    event UopenEvent(string ret);
    
    /**
     * Event for evaluate function.
     * @param ret - return message.
     * @param utility - address of utility with min distance.
     * @param index - index of load profile of utility with min distance.
     * @param allUtilities - addresses of all utilities (for demo reason only).
     * @param allDistances - distances of all utilities (for demo reason only).
     * @param allIndexes - indexes of load profiles of utilities (demo only).
     * @return void
     */
    event EvaluateEvent(string ret, address utility, uint index,
        address[] allUtilities, uint16[] allDistances, uint[] allIndexes);
    
    /**
     * Constructor runs only when the contract is created.
     * @param loadforecast - keccak256 hash of load forecast (32 bytes ascii).
     * @return void
     */
    function BestTariffVariant(bytes32 loadforecast) public {
        owner = msg.sender;
        loadforecasthash = loadforecast;
    }
    
    /**
     * Commit function for utilities to send their embedded load profile hashes.
     * @param profiles - array of keccak256 hashed embedded load profiles.
     * @return string
     */
    function commit(bytes32[] memory profiles) public returns (string ret) {
        // Check for existing load profile hashes.
        if (templatehashes[msg.sender].loadprofilehashes.length == 0) {
            for (uint i = 0; i < profiles.length; i++) {
                templatehashes[msg.sender].loadprofilehashes.push(profiles[i]);
            }
            ret = "Load profile hashes received.";
        } else {
            ret = "Load profile hashes already sent.";
        }
        // Prepare callback for dapp interface.
        CommitEvent(ret);
    }
    
    /**
     * Smart meter preparing load profile forecast commitment for evaluation.
     * @param embeddedForecast - bytes32 array of embedded forecast.
     * @param salt - bytes8 of random number used for hash calculation.
     * @return string
     */
    function smopen(bytes32[32] memory embeddedForecast, bytes8 salt) public
        returns (string ret)
    {
        // Check if message is from contract owner.
        if (msg.sender == owner) {
            // Check hashes and persist load forecast.
            if (getHash(embeddedForecast, salt) == loadforecasthash) {
                forecast = embeddedForecast;
                ret = "Load forecast commitment received.";
            } else {
                ret = "Invalid load forecast commitment.";
            }
        } else {
            ret = "Invalid sender.";
        }
        // Prepare callback for dapp interface.
        SmopenEvent(ret);
    }
    
    /**
     * Utility preparing embedded template load profiles for evaluation.
     * @param loadprofile - bytes32 array of embedded template load profile.
     * @param salt - bytes8 of random number used for hash calculation.
     * @return string
     */
    function uopen(bytes32[32] memory loadprofile, bytes8 salt) public
        returns (string ret)
    {
        // Check for existing forecast commitment.
        if (forecast.length > 0) {
            // Check for existing load profile commitment.
            if (templatehashes[msg.sender].loadprofilehashes.length > 0) {
                // Prepare hash for load profile check.
                bytes32 tmploadprofilehash = getHash(loadprofile, salt);
                ret = "Invalid load profile commitment.";
                for (uint i = 0;
                    i < templatehashes[msg.sender].loadprofilehashes.length;
                    i++)
                {
                    // Check hashes.
                    if (tmploadprofilehash == 
                        templatehashes[msg.sender].loadprofilehashes[i])
                    {
                        // Calculate distance and store result.
                        uint16 distance = getDistance(
                            forecast,
                            loadprofile
                        );
                        // Persist information.
                        TemplateDistance memory templDist;
                        templDist.utility = msg.sender;
                        templDist.distance = distance;
                        templDist.index = i;
                        templateDistances.push(templDist);
                        ret = "Load profile commitment received.";
                        break;
                    }
                }
            } else {
                ret = "Invalid sender, no load profile hashes committed.";
            }
        } else {
            ret = "Contract not opened for commitment.";
        }
        // Prepare callback for dapp interface.
        UopenEvent(ret);
    }
    
    /**
	 * Find load profile with smallest normalized Hamming distance to forecast.
     * @return string
     */
    function evaluate() public returns (string ret) {
        address utility;
        uint index;
        // Prepare return of all utilities for demo reasons only.
        address[] memory allU = new address[](templateDistances.length);
        uint16[] memory allD = new uint16[](templateDistances.length);
        uint[] memory allI = new uint[](templateDistances.length);
        // Evaluation is only allowed by contract owner.
        if (msg.sender == owner) {
            if (templateDistances.length > 0) {
                // Get minimum distance - maximum possible is 8192.
                uint16 min = 8193;
                for (uint i = 0; i < templateDistances.length; i++) {
                    if (templateDistances[i].distance < min) {
                        utility = templateDistances[i].utility;
                        min = templateDistances[i].distance;
                        index = templateDistances[i].index;
                    }
                    // Set values of all utilities for demo reasons only.
                    allU[i] = templateDistances[i].utility;
                    allD[i] = templateDistances[i].distance;
                    allI[i] = templateDistances[i].index;
                }
                ret = "Best tariff evaluated.";
            } else {
                ret = "No utility commitment.";
            }
        } else {
            ret = "Invalid sender.";
        }
        // Prepare callback for dapp interface.
        EvaluateEvent(ret, utility, index, allU, allD, allI);
    }
    
    /**
     * Generates hash of 8 elements of the embedding.
     * @param a - bytes32 of element a.
     * @param b - bytes32 of element b.
     * @param c - bytes32 of element c.
     * @param d - bytes32 of element d.
     * @param e - bytes32 of element e.
     * @param f - bytes32 of element f.
     * @param g - bytes32 of element g.
     * @param h - bytes32 of element h.
     * @return bytes32
     */
    function concat(bytes32 a, bytes32 b, bytes32 c, bytes32 d, bytes32 e,
        bytes32 f, bytes32 g, bytes32 h) internal pure returns (bytes32)
    {
        return keccak256(a, b, c, d, e, f, g, h);
    }
    
    /**
     * Generates hash of embedding and salt.
     * @param e - bytes32[32] array of embedding.
     * @param s - bytes8 salt.
     * @return bytes32
     */
    function getHash(bytes32[32] memory e, bytes8 s) internal pure
        returns (bytes32)
    {
        bytes32[4] memory elems;
        uint8 index = 0;
        for (uint8 i = 0; i < 32; i+=8) {
            elems[index] = concat(e[i], e[i+1], e[i+2],
                e[i+3], e[i+4], e[i+5], e[i+6], e[i+7]);
            index++;
        }
        return keccak256(elems[0], elems[1], elems[2], elems[3], s);
    }
    
    /**
     * Calculates Hamming distance of bytes32 arrays.
     * @param f - customer forecast embedding.
     * @param l - utility load profile embedding.
     * @return uint16
     */
    function getDistance(bytes32[32] memory f, bytes32[32] memory l) 
        internal view returns (uint16)
    {
        uint16 distance = 0;
        for (uint8 i = 0; i < 32; i++) {
            // Bitwise XOR over forecast and load profile.
            bytes32 xor = f[i] ^ l[i];
            for (uint8 ii = 0; ii < 32; ii++) {
                // Calculate distance of left and right 4 bits.
                distance += map[(uint8(xor[ii] & 0xf0)) >> 0x04];
                distance += map[uint8(xor[ii] & 0x0f)];
            }
        }
        return distance;
    }
}
